"""
A script to read a list of .spold files and obtain a NIS file with the following commands

* InterfaceTypes
* BareProcessors
* Interfaces

"""
import logging
import os

import math
import xlrd
from nexinfosys.command_generators.parser_ast_evaluators import get_nis_name
from nexinfosys.common.helper import download_file
from io import StringIO
import pandas as pd
from bw2io.importers.ecospold2 import Ecospold2DataExtractor
from enbios.common.helper import list_to_dataframe, generate_json, generate_workbook


def generate_csv(o):
    js = generate_json(o)
    df = pd.read_json(js)
    del df["activity"]
    del df["classifications"]
    del df["properties"]
    del df["loc"]
    del df["production volume"]
    del df["uncertainty type"]
    del df["amount"]  # We are generating a reference of Interface Types. Not specific to a particular Activity
    # TODO There are many duplicate "name"s, for now remove them
    df.drop_duplicates(subset=["name"], inplace=True)
    s = StringIO()
    s.write('# To regenerate this file, execute function "generate_csv" in script "read_lci_data (ENVIRO)"\n')
    df.to_csv(s, index=False)
    return s.getvalue()


def read_ecospold_file(f):
    if os.path.exists(f) and os.path.isfile(f):
        return Ecospold2DataExtractor.extract(f, "test", use_mp=False)
    else:
        return None


def lci_to_interfaces_csv(lci_file, csv_file):
    lci_tmp = read_ecospold_file(lci_file)
    s = generate_csv(lci_tmp[0]["exchanges"])
    with open(csv_file, "wt") as f:
        f.write(s)

"""
lci_tmp = read_ecospold_file("/home/rnebot/Downloads/Electricity, PV, 3kWp, multi-Si.spold")
lci_tmp = read_ecospold_file("/home/rnebot/Downloads/Electricity, PV, production mix.spold")
lci_tmp = read_ecospold_file("/home/rnebot/Downloads/5c5c2277-8df1-4a73-a479-9c14deec9bb1.spold")
s = generate_json(lci_tmp[0]["exchanges"])
with open("/home/rnebot/Downloads/exchanges.json.txt", "wt") as f:
    f.write(s)
s = generate_csv(lci_tmp[0]["exchanges"])
with open("/home/rnebot/Downloads/exchanges.csv", "wt") as f:
    f.write(s)

"""


class SpoldToNIS:
    def __init__(self):
        pass

    def _get_spold_files_from_correspondence_file(self, correspondence_path):
        df = pd.read_csv(correspondence_path, comment="#")
        # Assume all columns are present: "name", "match_target_type", "match_target", "weight", "match_conditions"
        spolds = []
        ecoinvent_filename_idx = None
        type_idx = None
        name_idx = None
        for idx, col in enumerate(df.columns):
            if col.lower() == "match_target":
                ecoinvent_filename_idx = idx
            elif col.lower() == "match_target_type":
                type_idx = idx
            elif col.lower() == "name":
                name_idx = idx
        if ecoinvent_filename_idx is not None and type_idx is not None and name_idx is not None:
            for idx, r in df.iterrows():
                ecoinvent_filename = r[df.columns[ecoinvent_filename_idx]]
                o_type = r[df.columns[type_idx]]
                name = r[df.columns[name_idx]]
                if ecoinvent_filename != "" and o_type != "" and name != "":
                    if o_type.lower() in ["lca", "lci"]:
                        spolds.append(dict(name=name, ecoinvent_filename=ecoinvent_filename))
        return spolds

    def _get_spold_files_from_nis_file(self, nis_url):
        bytes_io = download_file(nis_url)
        xl = pd.ExcelFile(bytes_io.getvalue(), engine='openpyxl')
        # xl = pd.ExcelFile(xlrd.open_workbook(file_contents=bytes_io.getvalue()), engine="xlrd")
        spolds = []
        for sheet_name in xl.sheet_names:
            if not sheet_name.lower().startswith("bareprocessors"):
                continue
            df = xl.parse(sheet_name, header=0)
            ecoinvent_filename_idx = None
            ecoinvent_name_idx = None
            name_idx = None
            for idx, col in enumerate(df.columns):
                if col.lower() == "@ecoinventfilename":
                    ecoinvent_filename_idx = idx
                elif col.lower() == "@ecoinventname":
                    ecoinvent_name_idx = idx
                elif col.lower() == "processor":
                    name_idx = idx
            if ecoinvent_filename_idx is not None and ecoinvent_name_idx is not None and name_idx is not None:
                for idx, r in df.iterrows():
                    ecoinvent_filename = r[df.columns[ecoinvent_filename_idx]]
                    ecoinvent_name = r[df.columns[ecoinvent_name_idx]]
                    name = r[df.columns[name_idx]]
                    if ecoinvent_filename != "" and not isinstance(ecoinvent_filename, float) and ecoinvent_name != "" and name != "":
                        spolds.append(dict(name=name, ecoinvent_filename=ecoinvent_filename, ecoinvent_name=ecoinvent_name))
        return spolds

    def spold2nis(self, default_output_interface: str, lci_base: str, correspondence, nis_base_url: str, output_file: str):
        """
        A method to transform Spold files into a Workbook with InterfacesTypes, BareProcessors and Interfaces

        :param default_output_interface: Name of the interface to use when no clear output interface is found in the .spold file
        :param lci_base: Local path where .spold files are located
        :param correspondence: Location of the correspondence file (open "sim_correspondence_example.csv" file)
        :param nis_base_url: Location of the NIS Base file, which can also have processors with LCI location
        :param nis_file: Path of the result workbook
        :return: None
        """

        # Read the list of Spold files to process
        spolds = []
        if correspondence:
            _ = self._get_spold_files_from_correspondence_file(correspondence)
            spolds.extend(_)
        if nis_base_url:
            _ = self._get_spold_files_from_nis_file(nis_base_url)
            spolds.extend(_)

        interface_types = {}
        processors = {}
        interfaces = {}

        # Process each .spold file
        for spold in spolds:
            file_name = f"{lci_base}{os.sep}{spold['ecoinvent_filename']}"
            lci = read_ecospold_file(file_name)
            if lci is None:
                logging.debug(f"Ecospold file '{file_name}' not found")
                continue
            lci = lci[0]
            # Obtain processor name
            p_name = lci["name"]
            # Processors
            processors[get_nis_name(p_name)] = dict(name=p_name, lca_code=lci["activity"],
                                                    lca_type=lci["activity type"], lca_file=lci["filename"],
                                                    description=lci["comment"])

            # Read exchanges (to produce InterfaceTypes and Interfaces)
            df = pd.read_json(generate_json(lci["exchanges"]))
            del df["activity"]
            del df["classifications"]
            del df["properties"]
            del df["loc"]
            del df["production volume"]
            del df["uncertainty type"]
            dfi = df.copy()

            # Interface Types
            del dfi["amount"]
            # There are duplicate "name"s, for now remove them
            dfi.drop_duplicates(subset=["name"], inplace=True)
            for idx, r in dfi.iterrows():
                sphere = "Technosphere" if r["type"].lower() != "biosphere" else "Biosphere"
                interface_types[get_nis_name(r["name"])] = dict(comment=r.get("comment", ""), flow=r["flow"], sphere=sphere, unit=r["unit"], lci_name=r["name"])
            # Interfaces
            tmp = df.groupby(['name']).sum()  # Acumulate (sum) repeated exchange names
            main_output = None
            for idx, r in tmp.iterrows():
                if r["amount"] == 1.0:
                    main_output = get_nis_name(idx)
            if main_output is None:
                main_output = default_output_interface
            # Add output interface first
            i_name = get_nis_name(main_output)
            interfaces[(p_name, i_name)] = dict(value=1, relative_to="", is_output=True)
            for idx, r in tmp.iterrows():
                i_name = get_nis_name(idx)
                if (p_name, i_name) not in interfaces:
                    relative_to = main_output if i_name != main_output else ""
                    interfaces[(p_name, i_name)] = dict(value=r["amount"], relative_to=relative_to,
                                                        is_output=i_name == main_output)

        # Generate the three commands, InterfaceTypes, BareProcessors, Interfaces
        cmds = []

        # InterfaceTypes
        lst = [["InterfaceTypeHierarchy", "InterfaceType", "Sphere", "RoegenType", "ParentInterfaceType", "Formula",
                "Description", "Unit", "OppositeSubsystemType", "Attributes", "@EcoinventName"]]
        if default_output_interface != "":
            lst.append(["lci", default_output_interface, "Technosphere", "Flow", "", "",
                        "Default-generic output for LCI activities which do not state explicitly its output interface",
                        "EJ", "", "", ""])
        for interface_type, props in interface_types.items():
            opposite = "Environment" if props["sphere"].lower() == "biosphere" else ""
            lst.append(["lci", interface_type, props["sphere"], "Flow", "", "",
                        props["comment"], props["unit"], opposite, "", props["lci_name"]])
        cmds.append(("InterfaceTypes", list_to_dataframe(lst)))

        # BareProcessors
        lst = [["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
                "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong",
                "Attributes", "@EcoinventName", "@EcoinventFilename", "@region"]]
        for processor, props in processors.items():
            spold_file = props["lca_file"]
            lst.append(["", processor, "", "", "", "Structural", "No", "", "", "", "", "", "", props["name"], spold_file, ""])
        cmds.append(("BareProcessors", list_to_dataframe(lst)))

        # Interfaces
        lst = [
            ["Processor", "InterfaceType", "Interface", "Sphere", "RoegenType", "Orientation", "OppositeSubsystemType",
             "GeolocationRef", "GeolocationCode", "InterfaceAttributes", "Value", "Unit", "RelativeTo", "Uncertainty",
             "Assessment", "PedigreeMatrix", "Pedigree", "Time", "Source", "NumberAttributes", "Comments"]]
        for p_i, props in interfaces.items():
            p_name = get_nis_name(p_i[0])
            v = props["value"]
            orientation = "Output" if props["is_output"] else "Input"
            output_name = props["relative_to"]
            lst.append([p_name, p_i[1], "", "", "", orientation, "", "", "", "", v,
                        "", output_name, "", "", "", "", "Year", "Ecoinvent", "", ""])
        cmds.append(("Interfaces", list_to_dataframe(lst)))

        s = generate_workbook(cmds)
        if s:
            with open(output_file, "wb") as f:
                f.write(s)
        else:
            print(f"ACHTUNG BITTE!: it was not possible to produce XLSX, probably because no .Spold file was found, "
                  f"check LCI data folder, {lci_base}, is correctly specified.")



import pandas as pd
from nexinfosys.embedded_nis import NIS
from typing import Dict
from enbios.input import Simulation
from enbios.input.lci import LCIIndex
from enbios.processing.nis_file_generator import generate_import_commands_command, generate_processors_command, \
    generate_interfaces_command


def block_meets_conditions(block, condition):
    """
    Parse expression with
    :param block:
    :param condition:
    :return:
    """
    # TODO arith_boolean_expression_with_less_tokens (DESPUÉS DE NEXINFOSYS V0.47)
    from nexinfosys.command_generators.parser_field_parsers import string_to_ast, arith_boolean_expression
    from nexinfosys.model_services import State
    from nexinfosys.command_generators.parser_ast_evaluators import ast_evaluator
    try:
        ast = string_to_ast(arith_boolean_expression, condition)
    except:
        print(f"A syntax error here")
    s = State()
    for k, v in block.items():
        s.set(k, v)
    issues = []
    res, unres = ast_evaluator(ast, s, None, issues)
    if isinstance(res, bool):
        return res
    else:
        return unres  # It is an error probably


class Matcher:
    """
    * Match Activities (LCA) and Technologies (simulation)
    * Match Technologies (simulation) into parent Processors (MuSIASEM)
    """
    def __init__(self, block_types_path, correspondence_path):
        self._block_types_path = block_types_path
        self._correspondence_path = correspondence_path
        self._typed_block = {}
        self._sim_lca = {}  # Store a list of target LCA activities for each simulation type
        self._sim_musiasem = {}  # Store a list of target MuSIASEM processors for each simulation type
        self._read()

    def _read(self):
        """ Read the correspondence file(s). Format:
technology,match_target_type,match_target,weight,match_conditions
wind_onshore_competing,lca,2fdb6da1-8281-453d-8a7c-0184dc3586c4_66c93e71-f32b-4591-901c-55395db5c132.spold,,
wind_onshore_competing,musiasem,Energy_system.Electricity_supply.Electricity_generation.Electricity_renewables.Electricity_wind.Electricity_wind_onshore,,
...
        """
        # ### Read block types (if defined) ###
        self._typed_block = {}
        df = pd.read_csv(self._block_types_path)
        # Assume all columns are present: "name", "type"
        for idx, r in df.iterrows():
            if self._typed_block.get(r["name"]):
                print(f"Entry for {r['name']} repeated. Overwriting")
            self._typed_block[r["name"]] = r["type"]

        # ### Read correspondence ###
        self._sim_lca = {}
        self._sim_musiasem = {}
        df = pd.read_csv(self._correspondence_path)
        # Assume all columns are present: "name", "match_target_type", "match_target", "weight", "match_conditions"
        for idx, r in df.iterrows():
            if r["match_target_type"].lower() == "lca":
                d = self._sim_lca
            else:
                d = self._sim_musiasem
            lst = d.get(r["name"], [])
            if len(lst) == 0:
                d[r["name"]] = lst
            lst.append(dict(target=r["match_target"], weight=r["weight"], conditions=r["match_conditions"]))

    def find_related_lci_activities(self, block: Dict[str, str], lci_index: LCIIndex):
        block_name = block["name"]
        lst = self._sim_lca.get(block_name)
        if not lst:
            block_type = self._typed_block.get(block)
            if block_type:
                return self.find_related_lci_activities(block, lci_index)
        else:
            lst2 = []
            for b in lst:
                if block_meets_conditions(block, b["match_conditions"]):
                    lst2.append(b)
            return [dict(activity=lci_index.read_activity(b["target"]), weight=b["weight"]) for b in lst2]

    def find_candidate_musiasem_parent_processors(self, nis: NIS, block):
        # TODO Find parent processors names in the correspondence, locate them in NIS and return the full hierarchy name
        #      of each
        pass

    def equivalent_interface_type_for_lci_flow(self, lca_flow: str, nis: NIS):
        pass

    def equivalent_interface_type_for_simulation_flow(self, simulation_flow: str, nis: NIS):
        pass


def merge_models(nis: NIS, matcher: Matcher, simulation: Simulation, lci_data_index: LCIIndex):
    """
    Inputs:
      - Assessment configuration
      - Base MuSIASEM NIS file
      - LCI data
      - Simulation
        - Structure
        - Inputs
        - Outputs
      - Correspondence file
    :return:
    """
    # TODO - Parse all inputs
    #        - Base MuSIASEM: read, have model
    #        - Simulation structure
    #        - Correspondence file:
    #          - NIS: every "leaf" processor is in the file
    #          - Simulation: every technology is in the file, and can be assigned to at least one processor (more than one?)
    #          - LCA: every technology has an LCI
    #      - Configuration: * SPACE. single system or system per location or per country?
    #                       * TIME. Single NIS or multiple iterations
    #      - Merge NIS file with Simulation and LCI data
    #      - Compute indicators. Data structure: a dataset (dimensions - values - attributes) for each multilevel, another for scalar
    #      - Generate: jupyter script Python-R with example graphs. Network graph?, Sankey?, Geolayer?
    blocks_as_processor_templates = []
    for block in simulation.blocks:
        # Information (interfaces) of block from the simulation
        # TODO Name, value, unit, input/output, time, location, source, etc.
        block_info = simulation.read_block_info(block)
        # Matching LCA information
        matching_lci_activities = matcher.find_related_lci_activities(block, lci_data_index)
        for lci_activity in matching_lci_activities:
            for activity_flow in lci_activity.flows:
                # TODO Name, value, unit, input/output, time, location, source, etc.
                block_info.update()
                # TODO Take note of "Impact Category" (analysis) -> Attribute "@ImpactCategory"

        # TODO Read "LCIA method" -> Schema of SUM (exchanges*weight) per Impact Category

        # TODO Convert input information into NIS InterfaceTypes (including change of scale?)
        block_info = matcher.convert_to_interface_types(block_info)
        # Matching MuSIASEM processors
        parents = matcher.find_candidate_musiasem_parent_processors(nis, block)
        # TODO obtain full processor names
        block_info["parents"] = parents
        # Append one more processor
        blocks_as_processor_templates.append(block_info)
    # Generate NIS commands
    lst = []
    lst.append(("Import", generate_import_commands_command([])))  # Input NIS file path
    # Enumerate the blocks and relate them to parent processors
    lst.append(("BareProcessors", generate_processors_command(blocks_as_processor_templates)))
    lst.append(("Interfaces", generate_interfaces_command(blocks_as_processor_templates)))
    # TODO Dump
    #  blocks with no MuSIASEM parents (as warning?),
    #  blocks with no matching LCI (as warning)
    #  flows with no matching InterfaceType (as error)
    return lst


# /usr/bin/env python
import os
import fire

from enbios.input.data_preparation.lci_to_nis import lci_to_interfaces_csv, SpoldToNIS
from enbios.input.data_preparation.recipe_to_nis import convert_recipe_to_nis
from enbios.input.data_preparation.sentinel_to_nis_prep import sentinel_to_prep_file
from enbios.processing.main import Enviro

"""
* Action to prepare a NIS file. Output issues (console), output state (to a file and hash)


"""

"""
DEVELOPMENT CLI EXECUTION:
 
python -m enbios.bin.script
  <it will show the automatically generated help>

CLI EXECUTION ("enbios" because it is defined with that name in setup.py):
 
enbios (same as previous)
  <it will show the -automatically generated- help>
  
"""


class EnbiosCLI:
    def recipe_to_nis(self, recipe_file: str, lcia_file: str):
        """
        Convert an XLSX file with ReCiPe2016 to a CSV file ready to be declared in a DatasetDef which can be imported
        in a "LCIAMethod" command

        Example:
           enbios recipe_to_nis /home/rnebot/GoogleDrive/AA_SENTINEL/workspace_20210331/ReCiPe2016_CFs_v1.1_20180117.xlsx /home/rnebot/GoogleDrive/AA_SENTINEL/case_studies/library/recipe2016_2.csv


        :param recipe_file: The full path of the input ReCiPe2016 XLSX file
        :param lcia_file: The full path of the output CSV file
        :return:
        """
        convert_recipe_to_nis(recipe_file, lcia_file)

    def sentinel_to_nis_preparatory(self, sentinel_data_package_json_path: str, nis_file: str):
        """
        Read a Sentinel Data Package and elaborate an XLSX file with NIS commands and complementary information
        describing what is inside the Sentinel Package (enumeration of Scenarios, Regions, etc.) and a template of
        CORRESPONDENCE file as one of the Worksheets.

        :param sentinel_data_package_json_path: The full path of the JSON file describing a Sentinel data package
        :param nis_file: Full path of the output XLSX NIS formatted file
        :return:
        """
        sentinel_to_prep_file(sentinel_data_package_json_path, nis_file)

    def lci_to_interface_types(self, lci_file_path: str, csv_file: str):
        """
        Extract the interface types from an LCI file

        :param lci_file_path: Path of a single .spold file
        :param csv_file: Output CSV with a listing of distinct interface types (potentially importable in InterfaceTypes)
        :return:
        """
        lci_to_interfaces_csv(lci_file_path, csv_file)

    def lci_to_nis(self, spold_files_folder: str, correspondence_path: str, nis_base_url: str, nis_structurals_base_path: str):
        """
        Scan correspondence file AND a NIS file to extract a NIS file with the commands:
           - InterfaceTypes
           - BareProcessors
           - Interfaces
        The correspondence file should be a CSV file with the following header:
           name,match_target_type,match_target,weight,match_conditions

        Example lines in this file:
           wind_onshore_competing,lca,2fdb6da1-8281-453d-8a7c-0184dc3586c4_66c93e71-f32b-4591-901c-55395db5c132.spold,,
           wind_onshore_competing,musiasem,Energy_system.Electricity_supply.Electricity_generation.Electricity_renewables.Electricity_wind.Electricity_wind_onshore,,

        Example:
           enbios lci_to_nis /home/rnebot/GoogleDrive/AA_SENTINEL/enviro_tmp/ /home/rnebot/GoogleDrive/AA_SENTINEL/enviro/sim_correspondence_example.csv https://docs.google.com/spreadsheets/d/15NNoP8VjC2jlhktT0A8Y0ljqOoTzgar8l42E5-IRD90/edit?usp=sharing /home/rnebot/Downloads/out.xlsx

        :param spold_files_folder: Local folder where .spold files are stored
        :param correspondence_path: Correspondence file
        :param nis_base_url: URL of a NIS file that will be used as Base for the assembly of the model
        :param nis_structurals_base_path: File name of the output NIS that will contain the structural processors
        :return:
        """
        s2n = SpoldToNIS()
        s2n.spold2nis("generic_energy_production", spold_files_folder, correspondence_path, nis_base_url, nis_structurals_base_path)

    def enviro(self, cfg_file_path):
        """
        The main function of the package, reads the contents of the configuration file, which can be either a JSON or YAML
        file. Following an example of YAML file contents:
            nis_file_location: "https://docs.google.com/spreadsheets/d/12AlJ0tdu2b-cfalNzLqFYfiC-hdDlIv1M1pTE3AfSWY/edit#gid=1986791705"
            correspondence_files_path: ""
            simulation_type: sentinel
            simulation_files_path: ""
            lci_data_locations:
              - ""
            output_directory: ""

        (you can copy and paste the example in an empty text file as reference, and ".yaml" as extension)

        :param cfg_file_path:
        :return:
        """
        t = Enviro()
        t.set_cfg_file_path(cfg_file_path)
        t.compute_indicators_from_base_and_simulation()


def main():
    os.environ["PAGER"] = "cat"
    fire.Fire(EnbiosCLI)


if __name__ == '__main__':
    main()



import os

import itertools
import tempfile

import pandas as pd
from typing import Tuple

from nexinfosys.common.decorators import deprecated
from nexinfosys.common.helper import download_file
from nexinfosys.embedded_nis import NIS
from nexinfosys.serialization import deserialize_state

from enbios.common.helper import generate_workbook, hash_array, prepare_base_state, list_to_dataframe
from enbios.input import Simulation
from enbios.input.lci import LCIIndex
from enbios.input.simulators.calliope import CalliopeSimulation
from enbios.input.simulators.sentinel import SentinelSimulation
from enbios.processing import read_parse_configuration, read_submit_solve_nis_file
from enbios.processing.model_merger import Matcher, merge_models


#####################################################
# MAIN ENTRY POINT  #################################
#####################################################
class Enviro:
    def __init__(self):
        self._cfg_file_path = None
        self._cfg = None

    def set_cfg_file_path(self, cfg_file_path):
        self._cfg = read_parse_configuration(cfg_file_path)
        self._cfg_file_path = cfg_file_path if isinstance(cfg_file_path, str) else None

    def _prepare_base(self, solve: bool):
        """

        :return:
        """
        return prepare_base_state(self._cfg["nis_file_location"], solve)

    def _get_simulation(self) -> Simulation:
        # Simulation
        if self._cfg["simulation_type"].lower() == "calliope":
            simulation = CalliopeSimulation(self._cfg["simulation_files_path"])
        elif self._cfg["simulation_type"].lower() == "sentinel":
            simulation = SentinelSimulation(self._cfg["simulation_files_path"])
        return simulation

    def _prepare_process(self) -> Tuple[NIS, LCIIndex, Simulation]:
        # Simulation
        simulation = self._get_simulation()
        # MuSIASEM (NIS)
        nis = read_submit_solve_nis_file(self._cfg["nis_file_location"], state=None, solve=False)
        # LCI index
        lci_data_index = LCIIndex(self._cfg["lci_data_locations"])

        return nis, lci_data_index, simulation

    def generate_matcher_templates(self, combine_countries=False) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Generate auxiliary DataFrames to help in the elaboration of:
          - correspondence file
          - block types file
        :param combine_countries:
        :return: A tuple with 4 pd.DataFrames
        """
        # Construct auxiliary models
        nis, lci_data_index, simulation = self._prepare_process()
        # TODO Build a DataFrame with a list of LCI title, to file code
        #   name,title,file
        lci_df = pd.DataFrame()
        # TODO Build a DataFrame with a list of BareProcessors
        #   processor,parent
        nis_df = pd.DataFrame()
        # TODO Build DataFrame "correspondence"
        #   name,match_target_type,match_target,weight,match_conditions
        #   OK   "lca"             ---          1      <generate>
        #  if combine_countries:
        #   for each <name> generate a line per country and a line without country
        correspondence_df = pd.DataFrame()
        # TODO Build DataFrame "block_types"
        #   name,type
        #   OK   <from an existing database of tech names to Calliope tech types>
        block_types_df = pd.DataFrame()
        return lci_df, nis_df, correspondence_df, block_types_df

    def _read_simulation_fragments(self, split_by_region=True, split_by_scenario=True, split_by_period=True):

        """
        Read simulation. Find fragments and then start an iteration on them
        Each fragment is made of the "processors" and their "interfaces"
        It is here where basic processors are built. Correspondence file has tech but it could also
        have carrier and region

        :return:
        """
        # Create simulation
        simulation = self._get_simulation()
        # Read Simulation, returning:
        # prd: PartialRetrievalDictionary
        # scenarios: list of scenarios
        # regions: list of regions
        # times: list of time periods
        # carriers: list of carriers
        # col_types: list of value (not index) fields
        # ct: list of pairs (tech, region)
        prd, scenarios, regions, times, techs, carriers, col_types, ct = simulation.read("")

        partition_lists = []
        if split_by_region and len(regions) > 0:
            partition_lists.append([("_g", r) for r in regions])
        if split_by_period and len(times) > 0:
            partition_lists.append([("_d", t) for t in times])
        if split_by_scenario and len(scenarios) > 0:
            partition_lists.append([("_s", s) for s in scenarios])

        for partition in list(itertools.product(*partition_lists)):
            partial_key = {t[0]: t[1] for t in partition}
            procs = prd.get(partial_key)
            # TODO Sweep "procs" and update periods, regions, scenarios, models and carriers
            md = dict(periods=[], regions=[], scenarios=[], models=[], carriers=[])
            yield partial_key, md, procs

    def compute_indicators_from_base_and_simulation(self):
        """
        MAIN entry point of current ENVIRO
        Previously, a Base NIS must have been prepared, see @_prepare_base

        :return:
        """
        # Prepare Base
        serial_state = self._prepare_base(solve=False)
        output_dir = tempfile.gettempdir()
        # Split in fragments, process each fragment separately
        # Acumulate results
        indicators = pd.DataFrame()
        for partial_key, fragment_metadata, fragment_processors in self._read_simulation_fragments():
            # fragment_metadata: dict with regions, years, scenarios in the fragment
            # fragment_processors: list of processors with their attributes which will be interfaces
            print(f"{partial_key}: {len(fragment_processors)}")
            results = process_fragment(serial_state, partial_key, fragment_metadata, fragment_processors, output_dir)

            # TODO Append "results" to "indicators"
            indicators += results

        # TODO Save indicators DataFrame as a single CSV file

    @deprecated  # Use "compute_indicators_from_base_and_simulation"
    def musiasem_indicators(self, system_per_country=True):
        """
        From the configuration file, read all inputs
          - Base NIS format file
          - Correspondence file
          - Simulation type, simulation location
          - LCI databases

        :param system_per_country:
        :return:
        """
        # Construct auxiliary models
        nis, lci_data_index, simulation = self._prepare_process()
        # Matcher
        matcher = Matcher(self._cfg["correspondence_files_path"])

        # ELABORATE MERGE NIS FILE
        lst = merge_models(nis, matcher, simulation, lci_data_index)

        # Generate NIS file into a temporary file
        temp_name = tempfile.NamedTemporaryFile(dir=self._cfg["output_directory"], delete=False)
        s = generate_workbook(lst)
        if s:
            with open(temp_name, "wb") as f:
                f.write(s)
        else:
            print(f"ACHTUNG BITTE!: it was not possible to produce XLSX")

        # Execute the NIS file
        nis, issues = read_submit_solve_nis_file(temp_name)

        # TODO Download outputs
        # TODO Elaborate indicators
        # TODO Write indicator files

        # os.remove(temp_name)


def process_fragment(base_serial_state, partial_key, fragment_metadata, fragment_processors, output_directory):
    state = deserialize_state(base_serial_state)
    # TODO List of -accounted- full processor names
    # TODO List of LCI processors

    processors = [
        ["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
         "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong", "Attributes",
         "@SimulationName", "@Region"]]  # TODO Maybe more fields to qualify the fragment
    interfaces = [
        ["Processor", "InterfaceType", "Interface", "Sphere", "RoegenType", "Orientation", "OppositeSubsystemType",
         "GeolocationRef", "GeolocationCode", "InterfaceAttributes", "Value", "Unit", "RelativeTo", "Uncertainty",
         "Assessment", "PedigreeMatrix", "Pedigree", "Time", "Source", "NumberAttributes", "Comments"]]
    clone = False
    if clone:
        cloners = [
            ["InvokingProcessor", "RequestedProcessor", "ScalingType", "InvokingInterface", "RequestedInterface",
             "Scale"]]
        clone_processors = [
            ["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural",
             "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong",
             "Attributes",
             "@SimulationName", "@Region"]]
        # TODO Find regions in the fragment
        #  for each region, create a processor in "clone_processors" and an entry in "cloners" hanging the top Processor into the top processor
        regions = []
        for r in regions:
            name = None
            clone_processors.append(["", name, "", "", "<system>", "Functional", "Yes", "",
                                     f"Country level processor, {t[1]}", "", "", "", "", t[1], t[1]])
            # TODO InvokingInterface, RequestedInterface, Scale are mandatory; specify something here
            cloners.append((name, "EnergySector", "Clone", "", "", ""))
    already_added_processors = set()
    for p in fragment_processors:
        # TODO Update time and scenario (can change from entry to entry, but for MuSIASEM it would be the same entity)
        time_ = None
        scenario = None
        # TODO Find MuSIASEM matches
        musiasem_matches = []  # find_musiasem_parents(p)
        first_name = None
        for m in musiasem_matches:
            name = None
            parent = None
            description = None
            original_name = None
            if first_name is None:
                first_name = ""
                if (name, parent) not in already_added_processors:
                    processors.append(["", name, parent, "", "system", "Structural", "Yes", "",
                                       description, "", "", "", "", original_name, t[1]])
                    already_added_processors.add((name, parent))
            else:
                if (name, parent) not in already_added_processors:
                    processors.append(["", name, parent, "", "system", "Structural", "Yes", "",
                                       "", "", "", "", "", "", ""])
                    already_added_processors.add((name, parent))

        if first_name is not None:
            for i in p.ifaces:
                interfaces.append([name, i, "", "Technosphere", "Flow", "<orientation>", "", "", "", "", v,
                                   "", "<relative_to>", "", "", "", "", time_, scenario, "", ""])

            lci_matches = []  # find_lci_matches(p)
            # TODO Find LCI matches, expand interfaces
            for cp in lci_matches:
                for i in cp.ifaces:
                    orientation = "Input"
                    v = 1  # TODO Value
                    relative_to = "flow_out"
                    interfaces.append([first_name, i, "", "", "", orientation, "", "", "", "", v,
                                       "", relative_to, "", "", "", "", time_, scenario, "", ""])

    # Generate NIS file
    cmds = []
    if clone:
        cmds.append(("BareProcessors regions", list_to_dataframe(clone_processors)))
        cmds.append(("ProcessorScalings", list_to_dataframe(cloners)))
    cmds.append(("BareProcessors simulation fragment", list_to_dataframe(processors)))
    cmds.append(("Interfaces simulation fragment", list_to_dataframe(interfaces)))
    s = generate_workbook(cmds)
    if s:
        temp_name = tempfile.NamedTemporaryFile(dir=output_directory, delete=False)
        with open(temp_name, "wb") as f:
            f.write(s)

        nis = NIS()
        nis.open_session(True, state)
        nis.load_workbook(temp_name)
        r = nis.submit_and_solve()
        tmp = nis.query_available_datasets()
        print(tmp)
        # TODO Obtain indicators matrix:
        #   (scenario, processor, region, time, ¿carrier?) -> (i1, ..., in)
        # TODO Append to global indicators matrix (this could be done sending results and another process
        #  would be in charge of assembling)
        nis.close_session()
        return None  # TODO Return a pd.DataFrame
    else:
        return pd.DataFrame()


if __name__ == '__main__':
    t = Enviro()
    _ = dict(nis_file_location="https://docs.google.com/spreadsheets/d/15NNoP8VjC2jlhktT0A8Y0ljqOoTzgar8l42E5-IRD90/edit#gid=839302943",
             correspondence_files_path="",
             simulation_type="sentinel",
             simulation_files_path="/home/rnebot/Downloads/borrame/calliope-output/datapackage.json",
             lci_data_locations={},
             output_directory="/home/rnebot/Downloads/borrame/enviro-output/")
    t.set_cfg_file_path(_)
    t.compute_indicators_from_base_and_simulation()
    print("Done")


